/*
 * $Header: /home/cvspublic/jakarta-tomcat/proposals/catalina/src/share/org/apache/tomcat/security/Attic/HttpBasicValve.java,v 1.5 2000/06/22 03:00:37 craigmcc Exp $
 * $Revision: 1.5 $
 * $Date: 2000/06/22 03:00:37 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */


package org.apache.tomcat.security;


import java.io.IOException;
import java.security.Principal;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.apache.tomcat.HttpRequest;
import org.apache.tomcat.HttpResponse;
import org.apache.tomcat.Realm;
import org.apache.tomcat.Session;
import org.apache.tomcat.deploy.LoginConfig;
import org.apache.tomcat.util.Base64;



/**
 * An <b>Authenticator</b> and <b>Valve</b> implementation of HTTP BASIC
 * Authentication, as outlined in RFC 2617:  "HTTP Authentication: Basic
 * and Digest Access Authentication."
 *
 * @author Craig R. McClanahan
 * @version $Revision: 1.5 $ $Date: 2000/06/22 03:00:37 $
 */

public final class HttpBasicValve
    extends HttpSecurityBase {


    // ----------------------------------------------------- Instance Variables


    /**
     * The Base64 helper object for this class.
     */
    private static final Base64 base64Helper = new Base64();


    /**
     * Descriptive information about this implementation.
     */
    private static final String info =
	"org.apache.tomcat.security.HttpSecurityValve/1.0";


    // ------------------------------------------------------------- Properties


    /**
     * Return descriptive information about this Valve implementation.
     */
    public String getInfo() {

	return (this.info);

    }


    // --------------------------------------------------------- Public Methods


    /**
     * Authenticate the user making this request, based on the specified
     * login configuration.  Return <code>true</code> if any specified
     * constraint has been satisfied, or <code>false</code> if we have
     * created a response challenge already.
     *
     * @param request Request we are processing
     * @param response Response we are creating
     * @param login Login configuration describing how authentication
     *              should be performed
     *
     * @exception IOException if an input/output error occurs
     */
    public boolean authenticate(HttpRequest request,
				HttpResponse response,
				LoginConfig config)
	throws IOException {

	// Have we already authenticated someone?
	Principal principal =
	    ((HttpServletRequest) request.getRequest()).getUserPrincipal();
	if (principal != null)
	    return (true);

	// Have we got a cached authenticated Principal?
	Session session = null;
	if (cache)
	    session = getSession(request);
	if (session != null) {
	    principal = session.getPrincipal();
	    if (principal != null) {
	        request.setAuthType("BASIC");
		request.setUserPrincipal(principal);
		return (true);
	    }
	}

	// Validate any credentials already included with this request
	HttpServletRequest hreq =
	    (HttpServletRequest) request.getRequest();
	HttpServletResponse hres =
	    (HttpServletResponse) response.getResponse();
	String authorization = request.getAuthorization();
	if (authorization != null) {
	    principal = findPrincipal(authorization, context.getRealm());
	    if (principal != null) {
	        request.setAuthType("BASIC");
		request.setUserPrincipal(principal);
		if (cache && (session != null))
		    session.setPrincipal(principal);
		return (true);
	    }
	}

	// Send an "unauthorized" response and an appropriate challenge
	String realmName = config.getRealmName();
	if (realmName == null)
	    realmName = hreq.getServerName() + ":" + hreq.getServerPort();
	hres.setHeader("WWW-Authenticate", 
                       "Basic realm=\"" + realmName + "\"");
        hres.setContentLength(0);
	hres.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
	hres.flushBuffer();
	return (false);

    }


    // -------------------------------------------------------- Private Methods


    /**
     * Parse the specified authorization credentials, and return the
     * associated Principal that these credentials authenticate (if any)
     * from the specified Realm.  If there is no such Principal, return
     * <code>null</code>.
     *
     * @param authorization Authorization credentials from this request
     * @param realm Realm used to authenticate Principals
     */
    private static Principal findPrincipal(String authorization, Realm realm) {

	// Validate the authorization credentials format
	if (authorization == null)
	    return (null);
	if (!authorization.startsWith("Basic "))
	    return (null);
	authorization = authorization.substring(6).trim();

	// Decode and parse the authorization credentials
	String unencoded =
	  new String(base64Helper.decode(authorization.getBytes()));
	int colon = unencoded.indexOf(':');
	if (colon < 0)
	    return (null);
	String username = unencoded.substring(0, colon).trim();
	String password = unencoded.substring(colon + 1).trim();

	// Validate these credentials in our associated realm
	return (realm.authenticate(username, password));

    }


}
